/*---------------------------------------------------------------------------------------
//Cache Class
 ----------------------------------------------------------------------------------------*/
 dtxFramework.Cache = function(framework, id) {
	dtxFramework.DTXElement.apply(this, [framework, id]);
	var self = this;
	
	//---------------------------------------------------------------------------------
	//Constants
	//---------------------------------------------------------------------------------
	this.$DEFAULT_TYPE = 'url';
	this.$DEFAULT_DIRECTORY = '\\';
	this.$DEFAULT_LIFE_TIME = 24*60*60*1000;
	this.$DEFAULT_CAPACITY = 20;
	this.$DEFAULT_GC_PROBABILITY = 0.1;
	this.$DEFAULT_INDEX_NAME = 'cache_index.xml';
	
	
	//---------------------------------------------------------------------------------
	//Properties
	//---------------------------------------------------------------------------------
	this.$type = null;							//Cache type: file or URL
	this.$directory = null;						//Directory associated with this cache
	this.$lifetime = null;						//Lifetime of cache items in minutes
	this.$capacity = null;						//Maximum number of items allowed in the cache
	this.$GCProbability = null;					//Garbage collection probabilty.
	this.$indexName = null;						//Cache index file
	this.$initialized = false;					//Set to true iff init() method was called
	
	//---------------------------------------------------------------------------------
	//Public  Methods
	//---------------------------------------------------------------------------------
	this.get = function(url) {
		if(!this.$initialized) {
			this.$init();
		}
		
		switch(this.$type) {
			case 'file':
				var matches = url.match(/^(.+)\/(.+)\?$/);
				if(matches) {
					var baseUrl = matches[1];
					var fileName = matches[2];
					
					if(!this.$isInCache(url)) {
						this.framework.download(url, this.$entryReady);
						this.$fireEvent('miss', {target: this, baseUrl: baseUrl, fileName: fileName});
					}
					else {
						this.$fireEvent('itemready', {target: this,  baseUrl: baseUrl, fileName: fileName});	
					}
				}
				break;
			case 'host':
			case 'url':
				if(!this.$isInCache(url)) {
					this.framework.download(url, this.$entryReady);
					this.$fireEvent('miss', {target: this, url: url});
				}
				else {
					var entryTag = this.$urlToEntryTag(url);
					var content = unescape(this.framework.loadFile(this.$directory + entryTag));
					this.$fireEvent('itemready', {target: this, url: url, content: content});
				}
				break;
		}
		
	}; 
	
	this.flush = function(url) {
		if(!this.$initialized)
			this.$init();
			
		var entryTag = this.$urlToEntryTag(url);
		var fileName;
		
		//Delete tag from index 
		var entries = this.$readIndex();
		if(entries) {
			var i = 0;
			while(i < entries.length)
				if(entries[i] == entryTag) {
					entries.splice(i);
					fileName = this.$directory + entryTag;
					this.framework.deleteFile(fileName);
					
					i = entries.length;
				}
				else {
					i++;
				}
				
			this.$saveIndex(entries);
		}
	};
	
	//---------------------------------------------------------------------------------
	// Private Methods
	//---------------------------------------------------------------------------------
	
	
	//Initializes the cache properties.
	this.$init = function() {	
		var cacheType = this.getAttribute('cachetype');
		
		if(typeof cacheType == 'string') {
			cacheType.toLowerCase();	
			if(cacheType == 'file' || cacheType == 'url' || cacheType == 'host') {
				this.$type = cacheType;
			}
			else {
				this.$type = this.$DEFAULT_TYPE;
			}
		}
		else {
			this.$type = this.$DEFAULT_TYPE;
		}
		
		
		var directory = this.getAttribute('directory');
		
		if(typeof directory == "string") {
			this.$directory = directory;
		}
		else {
			this.$directory = this.$DEFAULT_DIRECTORY;
		}
		
		var userProfileDir = this.framework.getUserProfileDir();
		this.framework.createDir(userProfileDir + this.$directory);	
		
		var lifetime = this.getAttribute('lifetime');
		if(typeof lifetime == 'string' && lifetime != '') {
			this.$lifetime = parseInt(lifetime, 10);
		}
		else {
			this.$lifetime = this.$DEFAULT_LIFETIME;
		}
		
		var capacity = this.getAttribute('capacity');
		if(typeof capacity == 'string' && capacity != '') {
			this.$capacity = parseInt(capacity, 10);
		}
		else {
			this.$capacity = this.$DEFAULT_CAPACITY;
		}	
		
		var indexName = this.getAttribute('indexname');
		if(typeof indexName == 'string' && indexName != '') {
			this.$indexName = indexName;
		}
		else {
			this.$indexName = this.$DEFAULT_INDEX_NAME;
		}
				
		if(!this.framework.fileExists(this.$directory + this.$indexName)) {
			var content = "<cacheindex></cacheindex>";
			this.framework.saveFile(this.$directory + this.$indexName, content); 
		}
		
		
		var gcProbability = this.getAttribute('gcprobability');
		if(typeof gcProbability == 'string' && gcProbability != '') {
			this.$gcProbability = parseFloat(gcProbability);
		}
		else {
			this.$gcProbability = this.$DEFAULT_GC_PROBABILITY;
		}
		
		this.$initialized = true;
	};
	
	
	this.$entryReady = function(url, code, content) {
		var entryTag = self.$urlToEntryTag(url);
		var cacheType = self.getAttribute('cachetype');
		
		if(cacheType == 'file') {
			var matches = url.match(/^(.+)\/(.+)\?$/);
			if (matches && mathches.length >= 3) {
				var baseUrl = matches[1];
				var fileName = matches[2];
			}
			else {
				throw "Invalid URL";
			}
		}
				
		if (code == 200) {
			switch(cacheType) {
			case 'file':
				self.framework.saveFile(self.$directory + fileName, escape(content));
				self.$fireEvent('itemready', {target: self, baseUrl: baseUrl, fileName: fileName});
			break;
			case 'host':
			case 'url':
				self.framework.saveFile(self.$directory + entryTag, escape(content));
				self.$fireEvent('itemready', {target: self, url: url, content: content});
			}
			self.$updateIndex(url);
			
		}
		else { 
			switch(cacheType) {
				case 'file':
					self.$fireEvent('noresponse', {target: self, baseUrl: baseUrl, fileName: fileName});
					break;
				case 'host':
				case 'url':
					self.$fireEvent('noresponse', {target: self, url: url});
			}
		}
	};
	
	// Returns true iff a cache entry has expired in the cache 
	this.$hasExpired = function(entryTag) {
		try {
			var fileName = this.$directory + entryTag;
			
			if(this.framework.fileExists(fileName)) {
				var lastModified = this.framework.getLastModifiedTime(fileName);
				var diff = new Date().getTime() - lastModified;
				return (diff > this.$lifetime ? true : false); 
			}
			else {
				return true;
			}
		} catch(e) {return true;}		
	};
	
	// Returns true iff a given entry is in cache
	this.$isInCache = function(url) {
		var entryTag = this.$urlToEntryTag(url);
		var entries = this.$readIndex();
		
		if(!entries) {
			return false;
		}
		
		var i = 0;
		while(i < entries.length && entries[i] != entryTag) {
			i++;
		}
				
		if(i == entries.length)
			return false;
		
		return !this.$hasExpired(entryTag); 
	};
	
	// Reads the cache's index
	// Returns an array of cache tags or null
	this.$readIndex = function() {
		try {
			var tags = new Array();
			var content = this.framework.loadFile(this.$directory + this.$indexName);
			var xmlDoc = this.framework.loadXML(content);
			if (xmlDoc == null)
			    return null;
			    
			var nodes = xmlDoc.getElementsByTagName('entry');
			
			if(nodes.length == 0)
				return null;
			
			for(var i = 0; i < nodes.length; i++) {
				tags.push(nodes[i].firstChild.nodeValue);
			}
			
			return tags;
		}
		catch(e) {
			return null;
		}
	};
	
	//Adds the entryTag to the cache index and performs garbage collection
	//Garbage collection strategy: last modified date is used as a first criterion, creation date as a second criterion
	this.$updateIndex = function(url) {
		var entryTag = this.$urlToEntryTag(url);
		var entries = this.$readIndex();

		if(!entries)
			entries = new Array();
		
		//Makes sure that tag is not already in cache before inserting it
		for(var i = 0 ; i < entries.length && entries[i] != entryTag ; i++);
		
		if(i == entries.length)
			entries.push(entryTag);
			
		//Expired entries are removed with probability GCProbability
		var freshEntries = new Array();
		var rnd = Math.random();
		if(rnd >= (1 - this.$GCProbability)) {
			for(var i = 0; i < entries.length; i++) {
				if(this.$hasExpired(entries[i])) {
					//alert(entries[i] + ' expired');
					this.framework.deleteFile(this.$directory + entries[i]);
				}
				else {
					freshEntries.push(entries[i]);
				}
			}
		}
		else {
			freshEntries = entries;
		}
		
		
		if(freshEntries.length > this.$capacity) {
			var toRemove = freshEntries.length - this.$capacity;
			while(toRemove > 0) {
				var entryTag = freshEntries.shift();
				this.framework.deleteFile(this.$directory + entryTag);
				toRemove--;
			} 
		}
		
		this.$saveIndex(freshEntries);
	};
	
	
	this.$saveIndex = function(entries) {	
		var xml = '<cacheindex>';
		for(var i = 0; i < entries.length; i++) {
			xml += '\n<entry>' + entries[i] + '</entry>';
		}
		xml += '\n</cacheindex>';
		this.$framework.saveFile(this.$directory + this.$indexName, xml);
	};
	
	this.$urlToEntryTag = function(url) {
		var cacheType = this.getAttribute('cachetype');
		
		switch(cacheType) {
			case 'host':
				var matches = url.match(/^http(s?):\/\/([\w.-]+)\/?/);
				if(matches) {
					url = matches[2];
				}
			case 'url':
				return (new MD5()).hex_md5(url);
			case 'file':
				var matches = url.match(/^(.+)\/(.+)\?$/);
				if (matches) {
					return matches[2];
				}
		}
	};
	
};

dtxFramework.inherits(dtxFramework.Cache, dtxFramework.DTXElement);


